#include <psptypes.h>
#include <pspkernel.h>
#include <pspsdk.h>
#include <pspsysclib.h>

#include <pspdisplay.h>
#include <pspge.h>




#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "libmenu.h"


#include "fonts/font_cg.c"

#ifdef LNGJPN
#include "fonts/misaki8x8.c"
#endif



#ifndef SCR_WIDTH
#define SCR_WIDTH 480
#endif

#ifndef SCR_HEIGHT
#define SCR_HEIGHT 272
#endif

#ifndef BUF_WIDTH
#define BUF_WIDTH 512
#endif




#define PB_CHAR_WIDTH			8
#define PB_CHAR_HEIGHT			8

#define	VRAM_SIZE				0x200000
#define	VRAM_ADDR_TOP			0x44000000


#define MF_MENU_MAX_NUMBER_OF_THREADS 64


/* PSP{^ */
#ifdef LNGJPN
	#define PB_SYM_PSP_UP			""
	#define PB_SYM_PSP_RIGHT		""
	#define PB_SYM_PSP_DOWN			""
	#define PB_SYM_PSP_LEFT			""
	#define PB_SYM_PSP_TRIANGLE		""
	#define PB_SYM_PSP_CIRCLE		""
	#define PB_SYM_PSP_CROSS		"~"
	#define PB_SYM_PSP_SQUARE		""
	#define PB_SYM_PSP_NOTE			""
#else
	#define PB_SYM_PSP_UP			"\x80"
	#define PB_SYM_PSP_RIGHT		"\x81"
	#define PB_SYM_PSP_DOWN			"\x82"
	#define PB_SYM_PSP_LEFT			"\x83"
	#define PB_SYM_PSP_TRIANGLE		"\x84"
	#define PB_SYM_PSP_CIRCLE		"\x85"
	#define PB_SYM_PSP_CROSS		"\x86"
	#define PB_SYM_PSP_SQUARE		"\x87"
	#define PB_SYM_PSP_NOTE			"\x88"
#endif


#define	PB_INVALID_FONT			('?' * PB_CHAR_HEIGHT)
#define	PB_INVALID_SJIS_INDEX	(0xFFFF)
#define PB_BLEND_FUNC( src_ch, dst_ch, alpha, maxbits ) ( ( ( ( src_ch ) + 1 - ( dst_ch ) ) * ( alpha ) >> ( maxbits ) ) + ( dst_ch ) )


#define PB_SWAP( a, b )		\
{							\
	int t = *(int *)a;		\
	*(int *)a = *(int *)b;	\
	*(int *)b = t;			\
}


#define PB_PUT_PIXEL( adr, color )														\
{																						\
	unsigned int pc = color;															\
	if( vram.colorConv ) pc = ( vram.colorConv )( pc, (void *)(vram.addr.buffer+adr) );	\
	memcpy( (void *)(vram.addr.buffer+adr), (void *)&pc, vram.pixelSize );				\
}


#define	IS_OPT(opt)			(opt && opt->type != SET_OPT_NONE)
#define	IS_OPT_FUNC(opt)	(opt && opt->type == SET_OPT_FUNC)
#define	IS_OPT_LIST(opt)	(opt && opt->type == SET_OPT_LIST)

#define	IS_OPT_FUNC_M(opt)	(IS_OPT_FUNC(opt) && opt->func.malloc_p)
#define	IS_OPT_FUNC_F(opt)	(IS_OPT_FUNC(opt) && opt->func.free_p)



typedef unsigned int ( *color_convert  )( unsigned int, void* );

typedef struct
{
	int				width;
	int				height;
	
	struct tagVRAM_ADDR
	{
		void		*display;
		void		*buffer;
	} addr;
	
	int				format;
	int				lineWidth;
	
	unsigned int	frameSize;
	unsigned int	lineSize;
	unsigned char	pixelSize;
	
	
	color_convert   colorConv;
} vram_frame_buffer;


typedef struct
{
	char   *buf;
	const size_t *len;
	size_t cpylen;
} xprintf_ctx;

static char menu_str[255];


static int thid_list_first[MF_MENU_MAX_NUMBER_OF_THREADS];
static int thid_count_first;

static int thid_list_Current[MF_MENU_MAX_NUMBER_OF_THREADS];
static int thid_count_Current;
	
static bool thid_status = false;
static bool thid_IsDump = false;
static bool this_IsBusy = false;


static vram_frame_buffer vram = {0},vram_tmp;
static bool blend = false;



void snprnt_cb( void *ctx, int ch )
{
	if( ch == 0x200 || ch == 0x201 ) return;

	xprintf_ctx *ctx_p = (xprintf_ctx*)ctx;
	
	//Buf͈͓̔ȂRs[
	if( ctx_p->cpylen  < *(ctx_p->len) )
	{
		ctx_p->buf[ctx_p->cpylen] = ch;
		ctx_p->cpylen++;
	}
}


int vsnprintf( char *buf, size_t n, const char *fmt, va_list ap )
{
	xprintf_ctx ctx = { buf, (const size_t *)&n, 0 };
	
	prnt( snprnt_cb, (void *)&ctx, fmt, ap );
	
	if( ctx.cpylen >= *(ctx.len) )
	{
		// BufTCY𒴂
		ctx.buf[*(ctx.len) - 1] = '\0';
	}
	else
	{
		// BufTCYɎ܂Ă
		ctx.buf[ctx.cpylen] = '\0';
	}
	
	return ctx.cpylen;
}

int snprintf( char *buf, size_t n, const char *fmt, ... )
{
	va_list ap;
	int ret;
	
	va_start( ap, fmt );
	ret = vsnprintf( buf, n, fmt, ap );
	va_end( ap );
	
	return ret;
}






static __inline u32 pb_color_8888_to_5650( unsigned int color, void *null )
{
	return ( ( ( color >> 3  ) & 0x1F ) | ( ( color >> 10 ) & 0x3F ) << 5 | ( ( color >> 19 ) & 0x1F ) << 11 );
}

static __inline u32 pb_color_8888_to_5551( unsigned int color, void *null )
{
	return ( ( ( color >> 3  ) & 0x1F ) | ( ( color >> 11 ) & 0x1F ) << 5 | ( ( color >> 19 ) & 0x1F ) << 10 | ( ( color >> 24 ) ? 1 : 0 ) << 15 );
}

static __inline u32 pb_color_8888_to_4444( unsigned int color, void *null )
{
	return ( ( ( color >> 4  ) & 0xF ) | ( ( color >> 12 ) & 0xF ) << 4 | ( ( color >> 20 ) & 0xF ) <<  8 | ( ( color >> 28 ) & 0xF ) << 12 );
}



static __inline unsigned int pb_color_blend_8888( unsigned int src, void *dst_addr )
{
	u8      alpha = ( src >> 24 ) & 0xFF;
	unsigned int dst   = *((u32 *)dst_addr);
	
	if( ! alpha )
	{
		return dst;
	}
	else if( alpha == 0xFF )
	{
		return src;
	}
	else
	{
		u32 color = alpha << 24;
		color |= PB_BLEND_FUNC( ( src       ) & 0xFF, ( dst       ) & 0xFF, alpha, 8 );
		color |= PB_BLEND_FUNC( ( src >>  8 ) & 0xFF, ( dst >>  8 ) & 0xFF, alpha, 8 ) << 8;
		color |= PB_BLEND_FUNC( ( src >> 16 ) & 0xFF, ( dst >> 16 ) & 0xFF, alpha, 8 ) << 16;
		return color;
	}
}

static __inline unsigned int pb_color_blend_4444( unsigned int src, void *dst_addr )
{
	u8  alpha = ( src >> 12 ) & 0xF;
	u16 dst   = *((u32 *)dst_addr);
	
	src = pb_color_8888_to_4444( src, NULL );
	
	if( ! alpha )
	{
		return dst;
	}
	else if( alpha == 0xF )
	{
		return src;
	}
	else
	{
		u16 color = alpha << 12;
		
		color |= PB_BLEND_FUNC( ( src      ) & 0xF, ( dst      ) & 0xF, alpha, 4 );
		color |= PB_BLEND_FUNC( ( src >> 4 ) & 0xF, ( dst >> 4 ) & 0xF, alpha, 4 ) << 4;
		color |= PB_BLEND_FUNC( ( src >> 8 ) & 0xF, ( dst >> 4 ) & 0xF, alpha, 4 ) << 8;
		
		return color;
	}
}

static __inline unsigned int pb_color_blend_5551( unsigned int src, void *dst_addr )
{
	u8  alpha = ( src >> 27 ) & 0x1F;
	u16 dst   = *((u32 *)dst_addr);
	
	src = pb_color_8888_to_5551( src, NULL );
	
	if( ! alpha )
	{
		return dst;
	}
	else if( alpha == 0x1F )
	{
		return src;
	}
	else
	{
		u16 color = ( alpha ? 1 : 0 ) << 15;
		
		color |= PB_BLEND_FUNC( ( src       ) & 0x1F, ( dst       ) & 0x1F, alpha, 5 );
		color |= PB_BLEND_FUNC( ( src >>  5 ) & 0x1F, ( dst >>  5 ) & 0x1F, alpha, 5 ) << 5;
		color |= PB_BLEND_FUNC( ( src >> 10 ) & 0x1F, ( dst >> 10 ) & 0x1F, alpha, 5 ) << 10;
		
		return color;
	}
}

static __inline unsigned int pb_color_blend_5650( unsigned int src, void *dst_addr )
{
	u8  alpha = ( src >> 26 ) & 0x3F;
	u16 dst   = *((u32 *)dst_addr);
	
	src = pb_color_8888_to_5650( src, NULL );
	
	if( ! alpha )
	{
		return dst;
	}
	else if( alpha == 0x3F )
	{
		return src;
	}
	else
	{
		u16 color = 0;
		
		color |= PB_BLEND_FUNC( ( src       ) & 0x1F, ( dst       ) & 0x1F, ( alpha >> 1 ) & 0x1F, 5 );
		color |= PB_BLEND_FUNC( ( src >>  5 ) & 0x3F, ( dst >>  5 ) & 0x3F, alpha                , 6 ) << 5;
		color |= PB_BLEND_FUNC( ( src >> 11 ) & 0x1F, ( dst >> 11 ) & 0x1F, ( alpha >> 1 ) & 0x1F, 5 ) << 11;
		
		return color;
	}
}





static __inline void *pb_buf_addr( int x, int y )
{
	return (void *)( ( x + y * vram.lineWidth ) * vram.pixelSize );
}


void libmSetBlend( bool state )
{
	blend = state;
}

bool libmGetBlend()
{
	return blend;
}

bool setMenuBuffers( int sync )
{
	vram.colorConv = NULL;
	int unk;
	
	sceDisplayGetMode( &unk, &vram.width, &vram.height);
	sceDisplayGetFrameBuf( &vram.addr.buffer, &vram.lineWidth, &vram.format, PSP_DISPLAY_SETBUF_NEXTFRAME );
	
	
	if( vram.addr.buffer && vram.lineWidth )
	{
		vram.addr.buffer= (void *)( (uintptr_t)vram.addr.buffer | VRAM_ADDR_TOP );
		
		vram.pixelSize	= ( vram.format == PSP_DISPLAY_PIXEL_FORMAT_8888 ? 4 : 2 );
		vram.lineSize	= vram.lineWidth * vram.pixelSize;
		vram.frameSize	= vram.lineSize * vram.height;
		
		
		//ʂ̍
		switch( vram.format )
		{
			case PSP_DISPLAY_PIXEL_FORMAT_4444:
			{
				vram.colorConv = ( blend ? pb_color_blend_4444 : pb_color_8888_to_4444 );
				break;
			}
			case PSP_DISPLAY_PIXEL_FORMAT_5551:
			{
				vram.colorConv = ( blend ? pb_color_blend_5551 : pb_color_8888_to_5551 );
				break;
			}
			case PSP_DISPLAY_PIXEL_FORMAT_565:
			{
				vram.colorConv = ( blend ? pb_color_blend_5650 : pb_color_8888_to_5650 );
				break;
			}
			
			case PSP_DISPLAY_PIXEL_FORMAT_8888:
			{
				if( blend ) vram.colorConv = pb_color_blend_8888;
				break;
			}
		}
		
		vram_tmp = vram;
		
		return true;
	}
	else
	{
		return false;
	}
}


void* getMenuBuffers()
{
	return vram.addr.buffer;
}

void swapMenuBuffers( int bufsync )
{
	if( vram.addr.display != vram.addr.buffer )
	{
		void *swap			= vram.addr.display;
		vram.addr.display	= vram.addr.buffer;
		vram.addr.buffer	= swap;
		
		sceDisplaySetFrameBuf( vram.addr.buffer , vram.lineWidth , vram.format , bufsync );
		sceDisplayWaitVblankStart();
		sceKernelDelayThread( 11500 );
	}
}


void clearMenuBuffers()
{
	memset(vram.addr.buffer ,0, vram.frameSize );
}

#ifdef LNGJPN
static __inline unsigned short pb_get_glyph_index_by_sjis( unsigned char sjis_hi, unsigned char sjis_lo )
{
	//SJIS ̖̈
	if ((sjis_hi < 0x81 || sjis_hi > 0xfc || sjis_lo < 0x40 || sjis_lo > 0xfc))	return PB_INVALID_SJIS_INDEX;
	
	unsigned char  kuten_hi = sjis_hi - 0x81;
	unsigned short kuten_lo = sjis_lo - 0x40;
	
	if( sjis_hi > 0x9F ) kuten_hi -= 0x40;
	if( sjis_lo > 0x7E ) kuten_lo--;
	
	kuten_hi <<= 1;
	if( kuten_lo > 0x5D ) kuten_hi++;
	
	kuten_lo %= 0x5E;
	
	if((kuten_hi >= 8 && kuten_hi <= 11) || (kuten_hi >= 13 && kuten_hi <= 14) || (kuten_hi >= 83 && kuten_lo > 5) || (kuten_hi > 83))
	{
		//tHg̖`̈
		return PB_INVALID_SJIS_INDEX;
	}
	else
	{
		/* 16ȍ~ł΁A`9`15敪 */
		if( kuten_hi > 7 && kuten_hi <=12 )
		{
			//9,10,11,12,
			kuten_hi -= 4;
			//kuten_lo--;
		}	
		else if( kuten_hi >= 15 )
		{
			//9,10,11,12,14,15
			kuten_hi -= 6;
			//kuten_lo--;
		}
	}
	
	//Ώەi[ĂAhX߂
	//  = kuten_hi() * 1̕(94) + kuten_lo * PB_CHAR_HEIGHT( = 8)
	return ((kuten_hi * 94 + kuten_lo) * PB_CHAR_HEIGHT);

}
#endif




int libmPrint( int x, int y, unsigned int fg, unsigned int bg, const char *str )
{
	uintptr_t      start_addr, put_addr;
	unsigned int   i;
	const char     *glyph;
	unsigned char  glyph_line_data, chr_width_bytes;
	unsigned short glyph_x, glyph_y;
	unsigned int   color;
	
	//if( !str || ( x < 0 || x > POS_X_MAX ) || ( y < 0 || y > POS_Y_MAX ) ) return 0;
	
	//VRAM̊JnAhX擾
	start_addr      = (uintptr_t)pb_buf_addr( x, y );
	put_addr        = start_addr;
	
	int cnt;
	
	for( i = 0,cnt = 0; str[i]; i++,cnt++ )
	{		
		//s
		if( (unsigned char)str[i] == '\n' )
		{
			start_addr += vram.lineSize * PB_CHAR_HEIGHT;
			put_addr = start_addr;
			continue;//Ɏ̕
		}
		
		/* Ot擾 */
		
#ifdef LNGJPN
		//{T|[g
		
		if( (unsigned char)str[i] < 0x80 )
		{
			// 0x80 < AXL[
			glyph = &(font_cg[(unsigned char)str[i] * PB_CHAR_HEIGHT]);
		}
		else if( (unsigned char)str[i] >= 0xA1 && (unsigned char)str[i] <= 0xDF )
		{
			//pJ^Ji
			glyph = &(font_hankaku_kana[((unsigned char)(str[i] - 0xA1) * PB_CHAR_HEIGHT)]);
		}
		else if( str[i + 1] )
		{
			//i2oCg)
			unsigned int  glyph_index = pb_get_glyph_index_by_sjis( str[i], str[i + 1] );
			i++;
			
			if( glyph_index != PB_INVALID_SJIS_INDEX )
			{	
				glyph = &(font_sjis[glyph_index]);
			}
			else
			{
				continue;
			}
		}
		else
		{
			//YȂ => ?
			continue;
		}
		
#else
		//p̂
		
		//AhX = R[h * i 1*8 = 8bytej
		if( (unsigned char)str[i] <= 0x88 )
		{
			glyph = &(font_cg[(unsigned char)str[i] * PB_CHAR_HEIGHT]);
		}
		else
		{
			//YȂ => ?
			continue;
		}
		
#endif
		
		chr_width_bytes = PB_CHAR_WIDTH * vram.pixelSize;
		
		for( glyph_y = 0; glyph_y < PB_CHAR_HEIGHT; glyph_y++, put_addr += vram.lineSize - chr_width_bytes )
		{
			glyph_line_data = glyph[glyph_y];
			
			for( glyph_x = 0; glyph_x < PB_CHAR_WIDTH; glyph_x++, glyph_line_data <<= 1, put_addr += vram.pixelSize )
			{
				color = glyph_line_data & 0x80 ? fg : bg;
				
				if( color != PB_TRANSPARENT ) PB_PUT_PIXEL( put_addr, color );
			}
		}
		
		put_addr -= vram.lineSize * PB_CHAR_HEIGHT - chr_width_bytes;
	}
	return cnt;
}



int libmPrintf( int x, int y, unsigned int fg, unsigned int bg, const char *format, ... )
{
	va_list ap;
	
	va_start( ap, format );
	vsnprintf( menu_str, 255, format, ap );
	va_end( ap );
	
	return libmPrint( x, y, fg, bg,  menu_str );
}



static void __inline pbLine( int sx, int sy, int ex, int ey, uint32_t color )
{
	uintptr_t draw_addr;
	int       e, dx, dy;
	
	if( color == PB_TRANSPARENT ) return;
	
	if( sx > ex ) PB_SWAP( &sx, &ex );
	if( sy > ey ) PB_SWAP( &sy, &ey );
	
	draw_addr = (uintptr_t)pb_buf_addr( sx, sy );
	
	if( sx == ex )
	{
		for( ; sy <= ey; sy++, draw_addr += vram.lineSize ) PB_PUT_PIXEL( draw_addr, color );
	}
	else if( sy == ey )
	{
		for( ; sx <= ex; sx++, draw_addr += vram.pixelSize ) PB_PUT_PIXEL( draw_addr, color );
	}
	else
	{
		e  = 0;
		dx = ex - sx;
		dy = ey - sy;
		
		if( dx > dy )
		{
			int x;
			
			for( x = sx; x <= ex; x++, draw_addr += vram.pixelSize )
			{
				e += dy;
				
				if( e > dx )
				{
					e -= dx;
					draw_addr += vram.lineSize;
				}
				
				PB_PUT_PIXEL( draw_addr, color );
			}
		}
		else
		{
			int y;
			
			for( y = sy; y <= ey; y++, draw_addr += vram.lineSize )
			{
				e += dx;
				
				if( e > dy )
				{
					e -= dy;
					draw_addr += vram.pixelSize;
				}
				
				PB_PUT_PIXEL( draw_addr, color );
			}
		}
	}
}



//j[\ɑXbh~
//vOCNȑOɗLdvXbh̃Z[tXg_v
bool dumpOtherThreads()
{
	thid_IsDump = ( sceKernelGetThreadmanIdList( SCE_KERNEL_TMID_Thread, thid_list_first, MF_MENU_MAX_NUMBER_OF_THREADS, &thid_count_first ) >= 0 );
	
	return thid_IsDump;
}




//vOCȊȎXbh~EĊJ
bool setThreadsStat( bool stat )
{
	if( !thid_IsDump || stat == thid_status || this_IsBusy ) return false;
	
	thid_status = stat;
	this_IsBusy = true;
	
	int i, j;
	int (*func)(SceUID) =  ( stat == MENU_SUSPEND ? sceKernelSuspendThread : sceKernelResumeThread );
	SceKernelThreadInfo thinfo;
	
	
	if( stat == MENU_SUSPEND )
	{
		//~
		
		int Current_id	= sceKernelGetThreadId();
		
		//݂̃XbhXg
		if( sceKernelGetThreadmanIdList( SCE_KERNEL_TMID_Thread, thid_list_Current, MF_MENU_MAX_NUMBER_OF_THREADS, &thid_count_Current ) < 0 )
		{
			//݂̃XbhXg擾s
			this_IsBusy = false;
			return false;
		}
		
		
		//݂̃XbhꗗƁAvOCNɃ_vZ[tXgr
		//Z[tXgƌXbhɈv̂͑ΏۊO
		for( i = 0; i < thid_count_Current; i++ )
		{
			memset(&thinfo, 0, sizeof(SceKernelThreadInfo));
			thinfo.size = sizeof(SceKernelThreadInfo);
			sceKernelReferThreadStatus(thid_list_Current[i], &thinfo);
			
			if
			(
			//	( strcasecmp( "SceImpose",thinfo.name) == 0 )
			//||
				( thinfo.attr & 0x100000 ) == 0x100000
			||
				( thid_list_Current[i] == Current_id )
			)
			{
				thid_list_Current[i] = 0;
				continue;
			}
			
			for( j = 0; j < thid_count_first; j++ )
			{
				if( thid_list_Current[i] == thid_list_first[j] )
				{
					thid_list_Current[i] = 0;
					break;
				}
			}
		}
	}
	
	
	//ŏIIȌ݂̃XbhXgɂXbh֒~EĊJ̑
	for( i = 0; i < thid_count_Current; i++ )
	{
		if( thid_list_Current[i] ) ( func )( thid_list_Current[i] );
	}
	
	this_IsBusy = false;
	
	return true;
}

bool getThreadsStat()
{
	return thid_status;
}


//ACeȕǂ
bool Is_Invalid_Item(MenuItem *Item)
{
	return ( !Item || Item->Type == Spacer || !Item->Visible || Item->Disable);
}




MenuItem* GetPrevItem( MenuItem *Item , bool Invalid_Skip )
{
	if( !Item || !Item->Prev ) return NULL;
	
	MenuItem *Ret	= NULL;
	MenuItem *Curr	= Item->Prev;
	
	while(Curr)
	{
		if( !Invalid_Skip || (Invalid_Skip && !Is_Invalid_Item(Curr)) )
		{
			Ret = Curr;
			break;
		}
		
		Curr = Curr->Prev;
	}
	
	return Ret;
}





MenuItem* GetNextItem( MenuItem *Item , bool Invalid_Skip )
{
	if( !Item || !Item->Next ) return NULL;
	
	MenuItem *Ret	= NULL;
	MenuItem *Curr	= Item->Next;
	
	while(Curr)
	{
		if( !Invalid_Skip || (Invalid_Skip && !Is_Invalid_Item(Curr)) )
		{
			Ret = Curr;
			break;
		}
		
		Curr = Curr->Next;
	}
	
	return Ret;
}



MenuItem* GetBottomItem( MenuItem *Item , bool Invalid_Skip )
{
	if( !Item || !Item->Next ) return NULL;
	
	MenuItem *Ret	= NULL;
	MenuItem *Curr	= Item;
	
	while(Curr)
	{
		Curr = GetNextItem(Curr,Invalid_Skip);
		
		if( Curr ) Ret = Curr;
	}
	
	return Ret;
}


MenuItem* GetTopItem( MenuItem *Item , bool Invalid_Skip )
{
	if( !Item || !Item->Prev ) return NULL;
	
	MenuItem *Ret 	= NULL;
	MenuItem *Curr	= Item;
	
	while( Curr )
	{
		Curr = GetPrevItem(Curr,Invalid_Skip);
		
		if( Curr )	Ret  = Curr;
	}
	
	return Ret;
}





/*
MenuItem* GetMenuItem(MenuContext* Context,int ID)
{
	MenuItem *Curr =  Context->Root ;
	
	while (Curr)
	{
		if (Curr->Children)
		{
			Curr = Curr->Children;
		}
		else
		{
			MenuItem* last = Curr;
			Curr = Curr->Next;
			
			if (!Curr && last->Parent)
			{
				Curr = last->Parent->Next;
			}
		}
	}
	
	return NULL;
}
*/






bool isMenuOpen(MenuContext* Context)
{
	if( !Context ) return false;
	
	return Context->IsOpen;
}


void MenuSetActiveItem( MenuContext *Context ,MenuItem *item )
{
	if( Context && item ) Context->Active = item;
}

MenuItem* MenuGetActiveItem(MenuContext* Context)
{
	if( !Context ) return NULL;
	
	return Context->Active;
}

bool MenuSetOpt( MenuContext* Context ,libmOpt *opt )
{
	if( !Context || !IS_OPT(opt) ) return false;
	
	Context->opt	= opt;
	opt->count		= 0;
	
	return true;
}

// Rootp̗̈m
MenuContext* createMenuContext(libmOpt *opt)
{
	MenuContext *context = NULL;
	
	if( !IS_OPT_FUNC_M(opt) ) return NULL;
	
	context = (MenuContext*)opt->func.malloc_p(sizeof(MenuContext));
	
	//Context쐬s
	if( !context ) return NULL;
	
	memset(context ,0,sizeof(MenuContext));
	
	//iftHgl ݒj
	context->HotKey.Show			= PSP_CTRL_NOTE ;
	context->HotKey.Back			= PSP_CTRL_CIRCLE;
	context->HotKey.Select			= PSP_CTRL_CROSS;
	context->HotKey.Up				= PSP_CTRL_UP;
	context->HotKey.Down			= PSP_CTRL_DOWN;
	context->HotKey.Left			= PSP_CTRL_LEFT;
	context->HotKey.Right			= PSP_CTRL_RIGHT;
	
	context->MenuInfo.StopThread	= false;
	
	context->MenuInfo.Type			= false;
	
	context->MenuInfo.AutoReturn	= true;
	
	context->Color.Active_Font		= GREEN;
	context->Color.Normal_Font		= WHITE;
	
	context->Color.Active_Back		= BLACK;
	context->Color.Normal_Back		= Invalid;
	
	context->Color.Disable_Font		= GRAY;
	context->Color.Disable_Back		= Invalid;
	
	context->MenuInfo.Lines		 	= 1;
	
	return context;
}

void removeMenuContext(MenuContext* Context)
{
	if( !Context || !IS_OPT_FUNC_F(Context->opt) ) return;
	
	MenuItem* Curr = Context->Root;
	
	
	//j[SACe폜(j
	while( Curr )
	{
		MenuItem *last = Curr;
		
		if ( Curr->Type == MenuContainer && Curr->Children )
		{
			Curr = Curr->Children;
		}
		else
		{
			Curr = Curr->Next;
			
			if (!Curr && last->Parent)
			{
				Curr = last->Parent->Next;
			}
		}
		
		Context->opt->func.free_p(last);
	}
	
	//ReLXg폜ij
	Context->opt->func.free_p(Context);
}

MenuItem* addMenuItem(MenuContext* Context, MenuItem* Parent, MenuItem* Item,u32 font_color,u32 back_color, int actionID, int Data)
{
	//Context ܂ Item oĂȂΉ
	if( !Context || !Item ) return NULL;
	
	MenuItem *Target = ( Parent ? Parent->Children : Context->Root );
	
	//lZbg
	Item->actionID		= actionID;
	Item->Data			= Data;
	Item->Parent		= Parent;
	
	Item->Color.Font	= font_color;
	Item->Color.Back	= back_color;
	
	if( Target )
	{
		MenuItem *Curr = Target;
		
		while (Curr->Next)
		{
			Curr = Curr->Next;
		}
		
		//ACȇOƌݒ
		Curr->Next = Item;
		Item->Prev = Curr;
	}
	else
	{
		//ACeȂ(܂ԍŏ)
		
		if( Parent )
		{
			Parent->Children	= Item;
		}
		else
		{
			Context->Root		= Item;
		}
		
		//ÕACe͂Ȃ
		Item->Prev = NULL;
		Item->Next = NULL;
	}
	
	return Item;
}

void removeMenuItem( MenuContext *Context , MenuItem* Item )
{
	if( !Context || !Item || !IS_OPT_FUNC_F(Context->opt) ) return;
	
	if( Item->Type == MenuContainer )
	{
		//폜ΏۂRei[ȂA̎qSč폜ij
		MenuItem* Curr = Item->Children;
		
		while( Curr )
		{
			MenuItem *last = Curr;
			
			if ( Curr->Type == MenuContainer && Curr->Children )
			{
				Curr = Curr->Children;
			}
			else
			{
				Curr = Curr->Next;
				
				if (!Curr && last->Parent)
				{
					Curr = last->Parent->Next;
				}
			}
			
			Context->opt->func.free_p(last);
		}
	}
	
	//폜Ώ Oƌ̃ACeȂ
	if( Item->Prev )
	{
		Item->Prev->Next = Item->Next;
		
		if( Item->Next )
		{
			Item->Next->Prev = Item->Prev;
		}
		else
		{
			Item->Prev->Next = NULL;
		}
	}
	else
	{
		Item->Prev = NULL;
	}
	
	//e̎q폜ΏۂȂ NULL 
	if( Item->Parent && Item->Parent->Children == Item ) Item->Parent->Children = NULL;
	
	//j[̃[g폜ΏۂȂ [gύX
	if( Item == Context->Root ) Context->Root = Item->Next;
	
	//ACe폜ij
	Context->opt->func.free_p(Item);
};


#define	GET_ARRAY_SIZE(size)	( size / sizeof(MenuItem*) )

static MenuItem* createItem(libmOpt *opt ,const char* Name,MenuItemType type)
{
	MenuItem* Item = NULL;
	
	
	if( IS_OPT_FUNC_M(opt) )
	{
		Item = (MenuItem*)opt->func.malloc_p(sizeof(MenuItem));
	}
	else if( IS_OPT_LIST(opt) && opt->count < GET_ARRAY_SIZE(opt->list.size) )
	{
		Item = opt->list.val + opt->count;
	}
	
	//Item쐬s
	if(!Item) return NULL;
	
	opt->count++;
	
	memset(Item,0,sizeof(MenuItem));
	
	//ŒKvȍڂݒ
	Item->Visible		= true;
	Item->Disable		= false;
	
	Item->Name		= Name;
	Item->Type		= type;
	
	Item->Color.Font	= Invalid;
	Item->Color.Back	= Invalid;
	Item->Color.Line	= Invalid;
	
	return Item;
}


// Rei쐬
MenuItem* createMenuContainer(libmOpt *opt , const char* Name )
{
	return createItem(opt,Name,MenuContainer);
}

MenuItem* createTriggerButton(libmOpt *opt , const char* Name )
{
	return createItem(opt,Name,TriggerButton);
}

MenuItem* createSpacer(libmOpt *opt , const char* Name )
{
	return createItem(opt,Name,Spacer);
}


MenuItem* createRadioButtonGroup(libmOpt *opt)
{
	return createItem(opt,NULL,RadioButtonGroup);
}


MenuItem* createRadioButton(libmOpt *opt , const char* Name, bool State ,MenuItem *Group )
{
	//eRei[ȂΒǉoȂ
	
	MenuItem* Item = createItem(opt,Name,RadioButton);
	
	//Item쐬s
	if( !Item || !Group || Group->Type != RadioButtonGroup ) return NULL;
	
	//`FbN
	Item->Ctrl.RB_Group = Group;
	
	//Iς݂݂̃ACe
	if( State ) Group->Ctrl.Group_Sel = Item;
	
	return Item;
}


MenuItem* createCheckBox(libmOpt *opt ,const char* Name, bool State)
{
	MenuItem* Item = createItem(opt,Name,CheckBox);
	
	//Item쐬s
	if(!Item) return NULL;
	
	Item->Ctrl.CB_Checked = State;
	
	return Item;
}




MenuItem* createSelectBox(libmOpt *opt ,const char* Name,const char *List[],int Selected)
{
	MenuItem* Item = createItem(opt,Name,SelectBox);
	
	//Item쐬s
	if(!Item) return NULL;
	
	//蔲ŁAIς݃ACe̓Xg͈͓̔`FbNĂȂ
	Item->Ctrl.SB.Index	= Selected;
	Item->Ctrl.SB.List	= List;
	
	return Item;
}

MenuItem* createUpDownCtrl(libmOpt *opt ,const char* Name,float Now,float Step,float Min, float Max,bool type)
{
	MenuItem* Item = createItem(opt,Name,UpDownCtrl);
	
	//Item쐬s
	if(!Item) return NULL;
	
	if(Now >= Min && Now <= Max)
	{
		//ݒlŏlȏ & őlȓ
		Item->Ctrl.UD.Now =  Now;
	}
	else
	{
		//ݒlŏlƍől͈̔͊O
		Item->Ctrl.UD.Now =  Min;
	}
	
	
	Item->Ctrl.UD.Step	= Step;
	Item->Ctrl.UD.Min	= Min;
	Item->Ctrl.UD.Max	= Max;
	
	Item->Ctrl.UD.type	= type;
	
	return Item;
}



void renderMenu(MenuContext* Context,int PosX,int PosY)
{
	if (Context && Context->IsOpen && Context->Root && vram.addr.buffer && vram.lineWidth )
	{
		u32 bg_color,font_color;
		
		int len ,x=0, y=0 , depth = 0 ,row = 0;
		
		MenuItem* Curr;
		
		
		if( Context->MenuInfo.Type == MENU_VIEW_ALL )
		{
			Curr = Context->Root;
		}
		else
		{
			Curr = ( Context->Active->Parent ? Context->Active->Parent->Children : Context->Root );
		}
		
		while( Curr )
		{
			len = 0;
			x = PosX + ((depth * 4) * PB_CHAR_WIDTH) ;
			y = PosY + (row  * (PB_CHAR_HEIGHT + Context->MenuInfo.Lines));
			
			//`ΏۂANeBuȃACeǂ
			//tHgƔwi̐Fς
			
			//ACe`F̗D揇
			//ANeBu(Context)>ACe>m[}(Context)
			
			if( Curr == Context->Active )
			{
				//ANeBuAeF̓VXeF
				font_color	= Context->Color.Active_Font;
				bg_color	= Context->Color.Active_Back;
			}
			else
			{
				//ANeBuł͂Ȃ
				
				if( Curr->Disable )
				{
					//ȃACeAVXe̖ACeF
					
					font_color	= Context->Color.Disable_Font;
					bg_color	= Context->Color.Disable_Back;
				}
				else
				{
					//tHgFAwiF͌ACew肪΂̐F
					//w肪Ȃ΃VXeF̃m[}J[
					
					font_color	= ( Curr->Color.Font != Invalid ? Curr->Color.Font : Context->Color.Normal_Font );
					bg_color	= ( Curr->Color.Back != Invalid ? Curr->Color.Back : Context->Color.Normal_Back );
				}
			}
			
			
			//Ԃ̃ACêݕ`
			if( Curr->Visible )
			{
				switch ( Curr->Type )
				{
					case RadioButton:
					{
						len = libmPrintf( x , y  , font_color, bg_color,"(%s)%s",((Curr->Ctrl.RB_Group && (Curr->Ctrl.RB_Group)->Ctrl.Group_Sel == Curr) ? "*" : " "), Curr->Name);
					}
					break;
					
					case CheckBox:
					{
						len = libmPrintf( x , y  , font_color, bg_color,"[%s]%s",Curr->Ctrl.CB_Checked ? "X" : " ", Curr->Name);
					}
					break;
					
					case MenuContainer:
					{
						bool flag = false;
						
						if(Curr->Ctrl.CNT.Disp == NULL)
						{
							len = libmPrintf( x,y, font_color, bg_color,"%s%s",Curr->Ctrl.CNT.IsOpen ? "-" : "+", Curr->Name);
						}
						else
						{
							if(Curr->Ctrl.CNT.Disp->Type == RadioButton)
							{
								if(Curr->Ctrl.CNT.Disp->Ctrl.RB_Group && (Curr->Ctrl.CNT.Disp->Ctrl.RB_Group)->Ctrl.Group_Sel->Visible)
								{
									len = libmPrintf( x,y, font_color, bg_color,"%s%s:\"%s\"",Curr->Ctrl.CNT.IsOpen ? "-" : "+", Curr->Name,(Curr->Ctrl.CNT.Disp->Ctrl.RB_Group)->Ctrl.Group_Sel->Name);
								}
								else
								{
									flag = true;
								}
							}
							else if(Curr->Ctrl.CNT.Disp->Type == SelectBox)
							{
								if(Curr->Ctrl.CNT.Disp->Visible)
								{
									len = libmPrintf( x,y, font_color, bg_color,"%s%s:\"%s\"",Curr->Ctrl.CNT.IsOpen ? "-" : "+", Curr->Name,Curr->Ctrl.CNT.Disp->Ctrl.SB.List[Curr->Ctrl.CNT.Disp->Ctrl.SB.Index]);
								}
								else
								{
									flag = true;
								}
							}
							else if(Curr->Ctrl.CNT.Disp->Type == CheckBox)
							{
								if(Curr->Ctrl.CNT.Disp->Visible)
								{
									len = libmPrintf( x,y, font_color, bg_color,"%s%s:[%s]",Curr->Ctrl.CNT.IsOpen ? "-" : "+", Curr->Name,Curr->Ctrl.CNT.Disp->Ctrl.CB_Checked ? "X" : " ");
								}
								else
								{
									flag = true;
								}
							}
							else if(Curr->Ctrl.CNT.Disp->Type == UpDownCtrl)
							{
								if(Curr->Ctrl.CNT.Disp->Visible)
								{
									if(!Curr->Ctrl.CNT.Disp->Ctrl.UD.type)
									{
										len = libmPrintf( x,y, font_color, bg_color,"%s%s:%d.%d",Curr->Ctrl.CNT.IsOpen ? "-" : "+", Curr->Name,(int)Curr->Ctrl.CNT.Disp->Ctrl.UD.Now,(int)((Curr->Ctrl.CNT.Disp->Ctrl.UD.Now*100)-(((int)Curr->Ctrl.CNT.Disp->Ctrl.UD.Now)*100)));
									}
									else
									{
										len = libmPrintf( x,y, font_color, bg_color,"%s%s:%d",Curr->Ctrl.CNT.IsOpen ? "-" : "+", Curr->Name,(int)Curr->Ctrl.CNT.Disp->Ctrl.UD.Now);
									}
								}
								else
								{
									flag = true;
								}
							}
							
							//֘AÂĂACeȏԂAʏ\
							if(flag)	len = libmPrintf( x,y, font_color, bg_color,"%s%s",Curr->Ctrl.CNT.IsOpen ? "-" : "+", Curr->Name);
						}
					}
					break;
					
					case SelectBox:
					{
						len = libmPrintf( x, y  , font_color, bg_color,"%s<%s[%s]%s>",
						Curr->Name ? Curr->Name : "",
						(Curr->Ctrl.SB.Index > 0 ? PB_SYM_PSP_LEFT : " "),
						Curr->Ctrl.SB.List[Curr->Ctrl.SB.Index],
						(Curr->Ctrl.SB.List[Curr->Ctrl.SB.Index+1] != NULL ? PB_SYM_PSP_RIGHT : " ")
						);
					}
					break;
					
					case UpDownCtrl:
					{
						if(!Curr->Ctrl.UD.type)
						{
							len = libmPrintf( x, y  , font_color, bg_color,"%s<-[%d.%d]+>",Curr->Name ? Curr->Name : "",(int)Curr->Ctrl.UD.Now,(int)((Curr->Ctrl.UD.Now*100)-(((int)Curr->Ctrl.UD.Now)*100)));
						}
						else
						{
							len = libmPrintf( x, y  , font_color, bg_color,"%s<-[%d]+>",Curr->Name ? Curr->Name : "",(int)Curr->Ctrl.UD.Now);
						}
					}
					break;
					
					case Spacer: 	if(Curr->Name == NULL) break;//Odefaultŕ\
					
					default:
					{
						len = libmPrint( x, y  , font_color, bg_color,Curr->Name);
					}
					break;
				}
				
				//LȂ`
				if(Context->MenuInfo.Lines != 0 && Curr->Color.Line != Invalid)
				{
					pbLine( x, y+PB_CHAR_HEIGHT, x+(PB_CHAR_WIDTH*len), y+PB_CHAR_HEIGHT, Curr->Color.Line);
				}
				
				row++;
			}
			
			
			//j[^Cvɂĕʏ
			if( Context->MenuInfo.Type == MENU_VIEW_ALL )
			{
				//^Cv0 S̕\
				//̃ACe`悷邽߂Ɉړ
				if ( Curr->Type == MenuContainer && Curr->Children &&  Curr->Ctrl.CNT.IsOpen )
				{
					Curr = Curr->Children;
					++depth;
				}
				else
				{
					MenuItem *last = Curr;
					Curr = Curr->Next;
					
					if (!Curr && last->Parent)
					{
						Curr = last->Parent->Next;
						--depth;
					}
				}
			}
			else
			{
				//^Cv1 ANeBuȃACê
				//Rei[݂̂
				Curr = Curr->Next;
			}
		}
		
		//if( thid_status == MENU_SUSPEND )
		//{
		//	//sceKernelDcacheWritebackAll();
		//}
	}
}



MenuParams* handleMenu(MenuContext* Context, SceCtrlData* Input)
{
	//ContextAANeBuȃACeȂ
	if( !Context || Context->Active == NULL ) return NULL;
	
	Context->Params.Item		= NULL;
	Context->Params.Action	= Menu_None;
	
	//OƈႤ{^Ă邩ǂ
	if( Context->LastState.Buttons != Input->Buttons )
	{
		//j[JĂARoot쐬ς
		if( Context->IsOpen && Context->Root )
		{
			if( (Input->Buttons & PSP_CTRL_HOME) == PSP_CTRL_HOME && thid_status == MENU_SUSPEND )
			{
				while(!setThreadsStat( MENU_RESUME ));
				
				Context->IsOpen = false;
			}
			else if( (Input->Buttons & Context->HotKey.Left) == Context->HotKey.Left )
			{
				//\L[̍ꂽ
				Context->Params.Action = Menu_Left;
				
				if( Context->Active->Type == SelectBox && Context->Active->Ctrl.SB.Index > 0 )
				{
					//Xg{bNX̃ACeO
					Context->Active->Ctrl.SB.Index--;
					Context->Params.Item = Context->Active;
				}
				else if( Context->Active->Type == UpDownCtrl )
				{
					//Abv_ẼACeSteplAZ
					if( (Context->Active->Ctrl.UD.Now - Context->Active->Ctrl.UD.Step) >= Context->Active->Ctrl.UD.Min )
					{
						Context->Active->Ctrl.UD.Now -= Context->Active->Ctrl.UD.Step;
					}
					
					Context->Params.Item = Context->Active;
				}
			}
			else if( (Input->Buttons & Context->HotKey.Right) == Context->HotKey.Right )
			{
				//\L[̉Eꂽ
				Context->Params.Action = Menu_Right;
				
				if( Context->Active->Type == SelectBox && Context->Active->Ctrl.SB.List[Context->Active->Ctrl.SB.Index+1] != NULL )
				{
					//Xg{bNX̃ACe
					Context->Active->Ctrl.SB.Index++;
					Context->Params.Item = Context->Active;
					
				}
				else if( Context->Active->Type == UpDownCtrl )
				{
					//Abv_ẼACeSteplAZ
					if( (Context->Active->Ctrl.UD.Now + Context->Active->Ctrl.UD.Step) <= Context->Active->Ctrl.UD.Max )
					{
						Context->Active->Ctrl.UD.Now += Context->Active->Ctrl.UD.Step;
					}
					
					Context->Params.Item = Context->Active;
				}
			}
			else if( (Input->Buttons & Context->HotKey.Up) == Context->HotKey.Up )
			{
				//\L[̏オꂽ
				Context->Params.Action = Menu_Up;
				
				MenuItem *Curr = GetPrevItem(Context->Active,true);
				
				if( !Curr && Context->MenuInfo.AutoReturn )
				{
					Curr = GetBottomItem(Context->Active,true);
				}
				
				if( Curr ) Context->Active = Curr;
			}
			else if( (Input->Buttons & Context->HotKey.Down) == Context->HotKey.Down )
			{
				//\L[̉ꂽ
				Context->Params.Action = Menu_Down;
				
				//ɗLȃACe邩ǂ
				MenuItem *Curr = GetNextItem(Context->Active,true);
				
				if( !Curr && Context->MenuInfo.AutoReturn )
				{
					//Rei[̍ŏֈCɖ߂
					Curr = GetTopItem(Context->Active,true);
				}
				
				if(Curr) Context->Active = Curr;
			}
			else if( (Input->Buttons & Context->HotKey.Select) == Context->HotKey.Select )
			{
				//L[̓
				Context->Params.Action = Menu_Select;
				
				//^Cvɂ蕪
				switch ( Context->Active->Type )
				{
					//Rei[
					case MenuContainer:
					{
						//q
						if ( Context->Active->Children )
						{
							//Ԃ̓ANeBuAqֈړ
							MenuItem  *Curr = Context->Active->Children;
							
							if( Is_Invalid_Item(Curr) )
							{
								Curr = GetNextItem(Curr,true);
							}
							
							
							//qɗLȃACeLǂ
							if( Curr )
							{
								//ANeBuȃACeq
								Context->Active->Ctrl.CNT.IsOpen	= true;
								Context->Active						= Curr;
							}
						}
						
						break;
					}
					
					//WI{^
					case RadioButton:
					{
						if ( Context->Active->Ctrl.RB_Group )
						{
							(Context->Active->Ctrl.RB_Group)->Ctrl.Group_Sel	= Context->Active;
							Context->Params.Item								= Context->Active;
						}
						
						break;
					}
					
					//`FbN{bNX
					case CheckBox:
					{
						//IԂ𔽓]
						Context->Active->Ctrl.CB_Checked	= !Context->Active->Ctrl.CB_Checked;
						Context->Params.Item				= Context->Active;
						
						break;
					}
					
					
					//gK[{^
					case TriggerButton:
					{
						//ꂽIDԋp
						Context->Params.Item = Context->Active;
						
						break;
					}
					
					default:
						break;
				}
			}
			else if( (Input->Buttons & Context->HotKey.Back) == Context->HotKey.Back )
			{
				if( Context->Active->Parent )
				{
					//߂A
					Context->Params.Action = Menu_Back;
					
					//̂ root ł͂ȂiRei[j
					
					Context->Active = Context->Active->Parent;
					Context->Active->Ctrl.CNT.IsOpen = false;
				}
				else
				{
					//j[ꂽ
					Context->Params.Action = Menu_Close;
					
					if( thid_status == MENU_SUSPEND )
					{
						sceDisplaySetFrameBuf( vram_tmp.addr.buffer , vram_tmp.lineWidth , vram_tmp.format , PSP_DISPLAY_SETBUF_IMMEDIATE);
						sceDisplayWaitVblankStart();
						sceDisplayWaitVblankStart();
						
						//sceKernelDcacheWritebackAll();
						
						//TXyhĂXbhĊJ
						while(!setThreadsStat( MENU_RESUME ));
					}
					
					Context->IsOpen = false;
				}
			}
		}
		else
		{
			if( (Input->Buttons & Context->HotKey.Show) == Context->HotKey.Show )
			{
				//j[\ꂽ
				Context->Params.Action = Menu_Show;
				
				if( Context->MenuInfo.StopThread )
				{
					//XbhTXyh
					while(!setThreadsStat( MENU_SUSPEND ));
					
					while(!setMenuBuffers(PSP_DISPLAY_SETBUF_NEXTFRAME));
					
					
					//_uobt@OpVRAMAhX
					if( (uintptr_t)vram.addr.buffer != VRAM_ADDR_TOP )
					{
						vram.addr.display = (void*)VRAM_ADDR_TOP;
					}
					else
					{
						vram.addr.display = (void*)( VRAM_ADDR_TOP + vram.frameSize);
					}
				}
				
				Context->IsOpen = true;
			}
		}
	}
	
	//݂Pad̂߂ɕێ
	Context->LastState = *Input;
	
	return &(Context->Params);
}

